# $Id: whiteboard.tcl,v 1.10 2004/03/22 21:52:22 aleksey Exp $

uplevel #0 [list source [file join [file dirname [info script]] svgrender.tcl]]

package require msgcat

set sysenc [encoding system]
encoding system utf-8
::msgcat::mcload [file join [file dirname [info script]] msgs]
encoding system $sysenc
unset sysenc

namespace eval wb {
    set id_base [pid]	;# used in proc create_id
}


proc wb::add_menu_items {m chatid} {
    $m add command -label [::msgcat::mc "Whiteboard"] \
	-command [list [namespace current]::open_wb $chatid]
}

hook::add chat_create_conference_menu_hook \
    [namespace current]::wb::add_menu_items
hook::add chat_create_user_menu_hook \
    [namespace current]::wb::add_menu_items


proc wb::add_roster_menu_item {m connid jids} {
    if {[info exists ::interface]} {
	ifacetk::roster::add_menu_item $m [::msgcat::mc "Whiteboard"] \
	    "[namespace current]::open_wb1 [list $connid] \$curuser" $jids
    } else {
	roster::add_menu_item $m [::msgcat::mc "Whiteboard"] \
	    "[namespace current]::open_wb1 [list $connid] \$curuser" $jids
    }
}

hook::add roster_create_user_menu_hook \
    [namespace current]::wb::add_roster_menu_item

proc wb::open_wb1 {connid jid} {
    open_wb [chat::chatid $connid $jid]
}

proc wb::open_wb {chatid} {

    set w [win_id whiteboard $chatid]
    if {[winfo exists $w]} {
	return
    }

    set jid [chat::get_jid $chatid]

    set title [format [::msgcat::mc "%s whiteboard"] $jid]
    add_win $w -title $title -tabtitle $title \
	-class Whiteboard \
	-raisecmd [list tab_set_updated $w]


    set sw [ScrolledWindow $w.sw]
    pack $sw -side right -fill both -expand yes
    set c [canvas $w.c -background white]
    $sw setwidget $c
    #pack $c -side right -fill both -expand yes

    set tb [frame $w.tb]
    pack $tb -side left -fill y

    set tbpolyline [radiobutton $w.tb.line -text [::msgcat::mc "PolyLine"] \
			-variable [namespace current]::tool($chatid) \
			-value polyline \
			-command [list [namespace current]::line_bind $c $chatid]]
    
    set tbfreehand [radiobutton $w.tb.freehand -text [::msgcat::mc "FreeHand"] \
			-variable [namespace current]::tool($chatid) \
			-value freehand \
			-command [list [namespace current]::freehand_bind \
				      $c $chatid]]

    set tbtext [radiobutton $w.tb.text -text [::msgcat::mc "Text"] \
		      -variable [namespace current]::tool($chatid) \
		      -value text \
		      -command [list [namespace current]::text_bind \
				    $c $chatid]]

    set tbmove [radiobutton $w.tb.move -text [::msgcat::mc "Move"] \
		    -variable [namespace current]::tool($chatid) \
		    -value move \
		    -command [list [namespace current]::move_bind \
				  $c $chatid]]

    set tbremove [radiobutton $w.tb.remove -text [::msgcat::mc "Remove"] \
		      -variable [namespace current]::tool($chatid) \
		      -value remove \
		      -command [list [namespace current]::remove_bind \
				    $c $chatid]]

    pack $tbpolyline $tbfreehand $tbtext $tbmove $tbremove -anchor w

    button $w.tb.clear -text [::msgcat::mc "Clear"] \
	-command [list [namespace current]::send_clear $chatid]
    pack $w.tb.clear -side bottom

    set text_set_fr [frame $w.tb.text_set_fr]
    set [namespace current]::text_set_fr($chatid) $text_set_fr
    label $text_set_fr.example_char -text A
    button $text_set_fr.seltextcol -text [::msgcat::mc "Text Color"] \
	-command [list [namespace current]::select_text_color \
		      $text_set_fr.seltextcol $text_set_fr.example_char]

    button $text_set_fr.selfont -text [::msgcat::mc "Font"] \
	-command [list [namespace current]::select_font \
		      $text_set_fr.selfont $text_set_fr.example_char]

    grid $text_set_fr.seltextcol $text_set_fr.selfont $text_set_fr.example_char
    pack $text_set_fr -side bottom

    canvas $w.tb.color -background \#000000 -height 0.5c -width 1
    button $w.tb.selcol -text [::msgcat::mc "Color"] \
	-command [list [namespace current]::select_color \
		      $w.tb.selcol $w.tb.color]
    pack $w.tb.selcol $w.tb.color -side bottom

    SpinBox $w.tb.width -width 4 -range {1 100} \
	-textvariable [namespace current]::width($chatid)
    pack $w.tb.width -side bottom

    trace variable [namespace current]::width($chatid) w \
	[list [namespace current]::change_width \
	     $w.tb.color [namespace current]::width($chatid)]

    variable tool
    set tool($chatid) polyline
    line_bind $c $chatid
}

proc wb::select_color {but col} {
    set color [SelectColor::menu $col.color [list below $but] \
		   -color [$col cget -background]]

    if {[string length $color]} {
        $col configure -background $color
    }
}

proc wb::select_font {chatid col} {
    variable app_font
    set font_desc [SelectFont .s -type dialog]
    if {![info exists app_font($font_desc)]} {
        set app_font($font_desc) [eval font create [list $font_desc] [font actual $font_desc]]
    }
    $col configure -font $app_font($font_desc)
}

proc wb::select_text_color {but col} {
    set color [SelectColor::menu $col.color [list below $but] \
		   -color [$col cget -foreground]]

    if {[string length $color]} {
        $col configure -foreground $color
    }
}

proc wb::get_text_color {chatid} {
    [set [namespace current]::text_set_fr($chatid)].example_char cget -foreground
}

proc wb::get_text_font {chatid} {
    [set [namespace current]::text_set_fr($chatid)].example_char cget -font
}

proc wb::get_color {chatid} {
    set w [win_id whiteboard $chatid]
    $w.tb.color cget -background
}

proc wb::change_width {col widthvar args} {
    set width [set $widthvar]
    $col configure -width $width
}

proc wb::get_width {chatid} {
    set w [win_id whiteboard $chatid]
    $w.tb.color cget -width
}

###############################################################################

proc wb::create_id {} {
    # Unseeded random number is not good enough, because remote clients
    # are likely to produce the same numbers.
    return [random 10000][clock seconds][set [namespace current]::id_base]
}

proc wb::send_svg {chatid tag} {
    if {[chat::is_groupchat $chatid]} {
	set type groupchat
    } else {
	set type chat
    }
    set connid [chat::get_connid $chatid]
    set jid [chat::get_jid $chatid]
    jlib::send_msg $jid \
	-connection $connid \
	-type $type \
	-xlist [list [jlib::wrapper:createtag x \
			  -vars {xmlns tkabber:whiteboard} \
			  -subtags [list [jlib::wrapper:createtag svg \
					      -subtags [list $tag]]]]]
}

proc wb::send_clear {chatid} {
    if {[chat::is_groupchat $chatid]} {
	set type groupchat
    } else {
	set type chat
        [win_id whiteboard $chatid].c delete all
    }
    set connid [chat::get_connid $chatid]
    set jid [chat::get_jid $chatid]
    jlib::send_msg $jid -type $type \
	-connection $connid \
	-xlist [list [jlib::wrapper:createtag x \
			  -vars {xmlns tkabber:whiteboard} \
			  -subtags [list [jlib::wrapper:createtag clear]]]]
}

proc wb::handle_wb {chatid from type body x} {
    foreach xelem $x {
	jlib::wrapper:splitxml $xelem tag vars isempty chdata children
	
	if {[cequal [jlib::wrapper:getattr $vars xmlns] tkabber:whiteboard]} {
	    open_wb $chatid
	    set w [win_id whiteboard $chatid]
	    foreach child $children {
		parse_item $chatid $child
	    }
	    tab_set_updated $w 1 message

	    $w.c configure -scrollregion [$w.c bbox all]
	}
    }
}
hook::add draw_message_hook [namespace current]::wb::handle_wb 1

proc wb::parse_item {chatid item} {
    set w [win_id whiteboard $chatid]
    jlib::wrapper:splitxml $item tag vars isempty chdata children

    switch -- $tag {
	svg {
	    foreach child $children {
		svg::parse_svg_item $w.c $child
	    }
	}
	move {
	    set id [jlib::wrapper:getattr $vars id]
	    set dx [jlib::wrapper:getattr $vars dx]
	    set dy [jlib::wrapper:getattr $vars dy]
	    if {![string is double $dx] || $dx == ""} {set dx 0}
	    if {![string is double $dy] || $dy == ""} {set dy 0}
	    $w.c move id$id $dx $dy
	}
	remove {
	    $w.c delete "id[jlib::wrapper:getattr $vars id]"
	}
	clear {
	    $w.c delete all
	}
    }
}

###############################################################################
# Line

proc wb::line_bind {c jid} {
    bind $c <ButtonPress-1> \
	[list [namespace current]::line_b1 [double% $c] [double% $jid] %x %y]
    bind $c <B1-Motion> {}
    bind $c <Motion> [list [namespace current]::line_b1m [double% $c] %x %y]
    bind $c <ButtonRelease-1> {}
    bind $c <Button-3> [list [namespace current]::line_b3 \
			    [double% $c] [double% $jid]]
}

proc wb::line_b1 {c jid x y} {
    variable line

    set x [$c canvasx $x]
    set y [$c canvasy $y]
    #puts "$x $y"

    if {[info exists line(drawed)]} {
	lappend line(coords) $x $y

	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords) $line(options)]
    } else {
	set line(drawed) 1
	set line(coords) "$x $y"
	set line(options) [list -fill [get_color $jid] \
			       -width [get_width $jid] \
			       -joinstyle miter]
    }
}

proc wb::line_b1m {c x y} {
    variable line

    set x [$c canvasx $x]
    set y [$c canvasy $y]

    if {[info exists line(drawed)]} {
	#lappend line(coords) $x $y

	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords) $x $y $line(options)]
    }
}


proc wb::line_b3 {c jid} {
    variable line

    catch {
	unset line(drawed)

	set id [create_id]
	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords) $line(options) \
			   -tag id$id]

	if {[chat::is_groupchat $jid]} {
	    $c delete $line(temp)
	}

	lappend vars points $line(coords)
	if {[set color [get_color $jid]] != "#000000"} {
	    lappend vars stroke $color
	}
	if {[set width [get_width $jid]] != 1} {
	    lappend vars stroke-width $width
	}

	lappend vars id $id

	send_svg $jid [jlib::wrapper:createtag polyline \
			   -vars $vars]

	set line(coords) {}
	set line(temp) {}
    }
}

###############################################################################

###############################################################################
# Freehand

proc wb::freehand_bind {c jid} {
    bind $c <ButtonPress-1> \
	[list [namespace current]::freehand_b1p \
	     [double% $c] [double% $jid] %x %y]
    bind $c <B1-Motion> [list [namespace current]::freehand_b1m \
			     [double% $c] %x %y]
    bind $c <ButtonRelease-1> \
	[list [namespace current]::freehand_b1r [double% $c] [double% $jid]]
    bind $c <Button-3> {}
}

proc wb::freehand_b1p {c jid x y} {
    variable line

    set x [$c canvasx $x]
    set y [$c canvasy $y]
    #puts "$x $y"

    if {0 && [info exists line(drawed)]} {
	lappend line(coords) $x $y

	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords)]
    } else {
	set line(drawed) 1
	set line(coords) "$x $y"
	set line(options) [list -fill [get_color $jid] \
			       -width [get_width $jid] \
			       -joinstyle round]
    }
}

proc wb::freehand_b1m {c x y} {
    variable line

    set x [$c canvasx $x]
    set y [$c canvasy $y]

    if {[info exists line(drawed)]} {
	lappend line(coords) $x $y

	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords) $x $y $line(options)]
    }
}


proc wb::freehand_b1r {c jid} {
    variable line

    catch {
	unset line(drawed)

	set id [create_id]
	catch {$c delete $line(temp)}
	set line(temp) [eval $c create line $line(coords) $line(options) \
			    -tag id$id]

	if {[chat::is_groupchat $jid]} {
	    $c delete $line(temp)
	}

	lappend vars points $line(coords)
	if {[set color [get_color $jid]] != "#000000"} {
	    lappend vars stroke $color
	}
	if {[set width [get_width $jid]] != 1} {
	    lappend vars stroke-width $width
	}

	lappend vars id $id

	send_svg $jid [jlib::wrapper:createtag polyline \
			   -vars $vars]

	set line(coords) {}
	set line(temp) {}
    }
}

###############################################################################

###############################################################################
# Remove

proc wb::remove_bind {c jid} {
    bind $c <ButtonPress-1> \
	[list [namespace current]::remove_b1p \
	     [double% $c] [double% $jid]]
    bind $c <B1-Motion> {}
    bind $c <ButtonRelease-1> {}
    bind $c <Button-3> {}
}

proc wb::remove_b1p {c chatid} {
    set tags [$c gettags current]
    set id ""
    foreach t $tags {
	if {[crange $t 0 1] == "id"} {
	    set id [crange $t 2 end]
	    break
	}
    }
    set connid [chat::get_connid $chatid]
    set jid [chat::get_jid $chatid]
    if {[chat::is_groupchat $chatid]} {
	set type groupchat
    } else {
	set type chat
	$c delete id$id
    }
    if {$id != ""} {
	jlib::send_msg $jid \
	    -connection $connid \
	    -type $type \
	    -xlist [list [jlib::wrapper:createtag x \
			      -vars {xmlns tkabber:whiteboard} \
			      -subtags [list [jlib::wrapper:createtag remove \
						  -vars [list id $id]]]]]
	
    }
}


###############################################################################

###############################################################################
# Move

proc wb::move_bind {c jid} {
    bind $c <ButtonPress-1> \
	[list [namespace current]::move_b1p \
	     [double% $c] [double% $jid] %x %y]
    bind $c <B1-Motion> [list [namespace current]::move_b1m \
			     [double% $c] %x %y]
    bind $c <ButtonRelease-1> \
	[list [namespace current]::move_b1r [double% $c] [double% $jid]]
    bind $c <Button-3> {}
}

proc wb::move_b1p {c jid x y} {
    variable move

    set tags [$c gettags current]
    set id ""
    foreach t $tags {
	if {[crange $t 0 1] == "id"} {
	    set id [crange $t 2 end]
	    break
	}
    }
    if {$id != ""} {
	set x [$c canvasx $x]
	set y [$c canvasy $y]

	set move(startx) $x
	set move(starty) $y
        set move(lastx) $x
	set move(lasty) $y

	set move(id) $id
    } else {
	catch {unset move(id)}
    }
}

proc wb::move_b1m {c x y} {
    variable move

    if {[info exists move(id)]} {
	set id $move(id)
	set x [$c canvasx $x]
	set y [$c canvasy $y]

	$c move id$id [expr {$x - $move(lastx)}] [expr {$y - $move(lasty)}]

	set move(lastx) $x
	set move(lasty) $y
    }
}


proc wb::move_b1r {c chatid} {
    variable move

    if {[info exists move(id)]} {
	set id $move(id)
	set x $move(lastx)
	set y $move(lasty)

	if {[chat::is_groupchat $chatid]} {
	    set type groupchat
	    $c move id$id \
		[expr {$move(startx) - $x}] [expr {$move(starty) - $y}]
	} else {
	    set type chat
	}

	set vars [list id $id \
		      dx [expr {$x - $move(startx)}] \
		      dy [expr {$y - $move(starty)}]]

	set connid [chat::get_connid $chatid]
	set jid [chat::get_jid $chatid]
	jlib::send_msg $jid \
	    -connection $connid \
	    -type $type \
	    -xlist [list [jlib::wrapper:createtag x \
			      -vars {xmlns tkabber:whiteboard} \
			      -subtags [list [jlib::wrapper:createtag move \
						  -vars $vars]]]]
    }
}

###############################################################################
# Text

proc wb::text_bind {c chatid} {
	bind $c <ButtonPress-1> {}
	bind $c <B1-Motion> {}
	bind $c <ButtonRelease-1> \
		[list [namespace current]::text_b1 [double% $c] [double% $chatid] %x %y]
	bind $c <Button-3> {}
}

proc wb::text_b1 {c chatid x y} {
	variable text_info
	set text_info(x) $x
	set text_info(y) $y
	set w [win_id whiteboard $chatid]
	if {[winfo exists $w.text_dialog]} {
		wm deiconify $w.text_dialog
	} else {
		toplevel $w.text_dialog
		entry $w.text_dialog.text -width 80 \
			-textvariable [namespace current]::text_entered($chatid)
		pack $w.text_dialog.text -side top
		button $w.text_dialog.okButton -text [::msgcat::mc "OK"] \
			-command [list [namespace current]::text_ok $c $chatid]
		button $w.text_dialog.cancelButton -text [::msgcat::mc "Cancel"] \
			-command [list wm withdraw $w.text_dialog]
		bind $w.text_dialog.text <Return> \
			[list [namespace current]::text_ok $c $chatid]
		pack $w.text_dialog.okButton -side left -anchor e
		pack $w.text_dialog.cancelButton -anchor w
	}
	focus $w.text_dialog.text
}

proc wb::text_ok {c chatid} {
	variable text_info
	set id [create_id]

	set text [set [namespace current]::text_entered($chatid)]
	if {[chat::is_groupchat $chatid]} {
		set type groupchat
	} else {
		variable app_font
		set type chat
		$c create text $text_info(x) $text_info(y) -tag id$id \
			-text $text \
			-fill [get_text_color $chatid]
		set font [get_text_font $chatid]
		if {[info exists app_font($font)]} {
			$c itemconfigure id$id -font $font
		}
	}
	set vars [list id $id x $text_info(x) y $text_info(y) \
		fill [get_text_color $chatid]]
	set font [get_text_font $chatid]
	if {[info exists app_font($font)]} {
		array set font_opt [font configure $font]
		lappend vars font-size $font_opt(-size) \
			font-family $font_opt(-family)
		if {$font_opt(-underline) || $font_opt(-overstrike)} {
			set dec {}
			if {$font_opt(-underline)} {
				lappend dec underline
			}
			if {$font_opt(-overstrike)} {
				lappend dec line-through
			}
			lappend vars text-decoration $dec
		}
		if {[string equal $font_opt(-slant) italic]} {
			lappend vars font-style italic
		}
		if {[string equal $font_opt(-weight) bold]} {
			lappend vars font-weight bold
		}
		unset font_opt
	}
	send_svg $chatid [jlib::wrapper:createtag text \
			   -vars $vars -chdata $text]
	set w [win_id whiteboard $chatid]
	wm withdraw $w.text_dialog
}


###############################################################################

